/*

 Simple XPressNet Throttle
 
 Version 1.1
 
 Copyright (C) 2010 John Alexander Stewart
 Con las correcciones y algún comentario más de Norber (C) 2015
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 Notes:
 
 1) There can be output on the serial console, so if running this from the 
 Arduino development GUI, start up the "console".
 
 2) This file contains lots of prints for possible XPressNet messages. These
 are here but (so far) are not called. They will eventually be removed, once
 the software is shown to be stable.
 
 3) This file HARD CODES in both the XPressNet 3.6 bus address, and the DCC
 decoder target address.
 
 4) It expects headlight is at F0; while dimming not currently supported. Headlight
 is turned on/off automatically when movement is detected. See the constants below
 for the delay between last time the locomotive moved and when the headlight is 
 turned off.
 
 5) Here are the hardware connections on the Arduino:
 
 Pin 2 - control of MAX485 - send or receive control for XPressNet.
 Pin 13 - LED - solid on for ok; flashing for "track power off"
 (see more possible messages below)
 Pin A0 - Analog input pin from potentiometer
 Pin A1 - "forward" or "off" from SPDT switch (forward == grounded)
 Pin A2 - "reverse" or "off" from SPDT switch (reverse == grounded)
 GND   - to the throttle common ground, and to the XPressNet "M" connection.
 VCC   - the Arduino will provide this - it is +5V.
 Not Connected   - voltage input from the XPressNet "L" connection. +12v typical.
 
 6) and on the MAX485 -
 
 Pin 1  RO Serial rx   -   to Arduino Serial receive Pin "RX<-0"
 Pin 2  /RE rec ena     -  to Arduino Pin 2
 Pin 3  DE send ena     -  to Arduino Pin 2
 Pin 4  DI serial tx    -  to Arduino Serial transmit Pin "TX->1"
 Pin 5  GND             -  to Arduino ground
 Pin 6  A               -  to XPressNet connector "A" 
 Pin 7  B               -  to XPressEnt connector "B"
 Pin 8  VCC             -  to +5V cc supplied from the Arduino.
 
 7) 2 inputs  as outlined in the Arduino connections above; one is a SPDT (centre
 off) switch that alternatively grounds either pin 2 or pin 3 for forward/rev/
 neutral; if it is a "typical" switch with 3 contacts on the back, the center
 contact will be the "ground" connection.
 
 a Linear taper potentiometer; really any value from 10kΩ above should work fine,
 one side tied to ground, the other to Vcc (+5V) and the wiper to analogue pin A0.
 
 8) We communicate at 62.5k, 1 start bit, 9 data bits, 1 stop bit, no parity. Note that
 the receive data drives the throttle; there are no other timers or interrupts. 
 Without the serial receive, nothing happens.
 
 -----------------------------------------------------------------------------------
 Initialization:
 
 At power up, we go through an initialization step:
 
 1)  GET_COMMAND_STATION_STATUS_STEP 
 (ask the command station for its current status)
 
 1a)  rinse and repeat step 1) until we are powered up.
 
 2)  GET_COMMAND_STATION_SW_VERSION_STEP
 (ask the command station for its software version)
 
 Ensure that we are at V3.0 or above.
 
 3)  GET_LOCOMOTIVE_INFO_STEP
 (we find out our locomotive speed steps, current direction and
 speed, etc)
 
 then either, if we are ok: PAST_INITIALIZATION
 if we are in a MU: LOCOMOTIVE_NOT_FREE
 and flash the LED quickly.
 
 
 LED Indications:
 
 Solid on:            Everything ok
 Slow 50/50 flashing: Track power off OR INVALID XBUS ADDRESS
 1 quick flash:       Locomotive part of a double header
 2 quick flashes:     Locomotive not free - MU, not single locomotive
 3 quick flashes:     Locomotive not free - part of a MU.
 4 quick flashes:     Speed steps not supported - not 14,27,28,128.
 5 quick flashes:     Command station not XPressNet 3.0 or higher
 6 quick flashes:     internal error - Speed step not supported, but should have been caught before here.
 
 */

#include <SoftwareSerial.h>
SoftwareSerial Debug(0, 1); // RX, TX
#define VERBOSE false   //When FALSE most messages to computer will not be sent

// This is our loco id, in decimal format. It will get split into bytes
// later, for sending to the system.

#define LOCO_ID 13    // NO leading zeroes. This is the address programmed into
// the decoder. A leading zero here will try and make the
// address as an OCTAL (base 8) address. Don't do this,
// as nobody in their right mind would keep track of 
// locomotives in octal format. Hex or binary maybe, but not
// octal.

// XpressNet address: must be in range of 1-31; must be unique. Note that some IDs
// are currently used by default, like 2 for a LH90 or LH100 out of the box, or 30
// for PC interface devices like the XnTCP.
#define MY_ADDRESS 5

// We will turn off the headlight (F0) after so many milliseconds.
// Light off timer starts when we go into neutral, or to zero speed.
// Light automatically on when going into gear (forward or reverse) and
// when we go from speed zero to some speed.
// 1 second, use 1000L, 5 seconds, 5000L, 10 seconds 10000L; basically
// we put tree zeroes and an L on the end of the number of seconds.
#define LIGHT_OFF_DELAY 2000L

// control of the XPressNet RS-485 transmission. 
#define MAX485_CONTROL 2          //Arduino Pin for send or receive control
#define LISTEN_MODE digitalWrite (MAX485_CONTROL, LOW)
#define SEND_MODE digitalWrite (MAX485_CONTROL, HIGH)

// status light.
#define LED_PIN 13
#define LED_ON digitalWrite (LED_PIN,HIGH);
#define LED_OFF digitalWrite (LED_PIN,LOW);

// where the potentiometer wiper wire goes:
#define POT_PIN A0

// for the forward/reverse switches = digital input pins
#define FORWARD_SWITCH A1
#define REVERSE_SWITCH A2

// Which serial port is used, if we have more than one on the chip?
// note that the 328s (the currently produced "smaller" chips) only
// have one serial port, so we force this.
#ifdef __AVR_ATmega328P__
#define SERIAL_PORT_0
#undef SERIAL_PORT_1
#else
//Maybe we are running on a MEGA chip with more than 1 serial port? If so, you
//can put the Xbus serial port to port 1, and use the port 0 for status messages
//to your PC. Good for debugging!
#define SERIAL_PORT_1
#undef SERIAL_PORT_0
#endif

// when sending data, do NOT continue until the hardware has sent the data out

#ifdef __AVR_ATmega8__
#define WAIT_FOR_XMIT_COMPLETE {while (!(UCSRA & (1<<TXC))); UCSRA = (1<<TXC); UCSRA = 0;}
#else
#ifdef SERIAL_PORT_0
#define WAIT_FOR_XMIT_COMPLETE {while (!(UCSR0A & (1<<TXC0))); UCSR0A = (1<<TXC0); UCSR0A = 0;}
#else
#define WAIT_FOR_XMIT_COMPLETE {while (!(UCSR1A & (1<<TXC1))); UCSR1A = (1<<TXC1); UCSR1A = 0;}
#endif

#endif

// XPressnet Call Bytes.
// Broadcast to everyone, we save the incoming data and process it later.
#define GENERAL_BROADCAST 0x160

// the address we look for when we are listening for ops to ME - 0x40 + GA.
// again, save the incoming data and process it later.
unsigned int myDirectedOps;

// the address we look for for our Call Byte Window - 0x60 + GA. When
// this sucker comes in, we are able to send data to the XPressNet bus.
unsigned int myCallByteInquiry;

// the address for a request acknowlegement sent from the stn - 0x00 + GA
// if this one comes in, we respond immediately with a "we are here" message.
unsigned int myRequestAck;

// we save incoming data for decoding. I think that the largest packet is 6-7 bytes. 
// If we save 20 bytes, we should be more than fine.
#define MAXSAVED 20

// for saving incoming data, we use these 2 variables.
int dataSaved;
int savedData[MAXSAVED];

// certain global XPressnet status indicators
boolean emergencyStop = false;
boolean trackPowerOff = false;
boolean inServiceMode = false;

// for the status LED;
int displayLEDState = LOW;    
boolean displayLEDflashing = false;
long flashInterval = 300;
long currentFlashCounter = 0;

// for the analogue potentiometer
int potValue = 0;                    // as read from the hardware
int oldPotValue = 2000;              // send ONLY changes
unsigned int analogueReadCounter = 0;// only do this every so often

int decoderSpeedSteps = 0;          // 28 bits by default on decoders...

// for the forward/reverse switches
boolean forward_closed = false;
boolean reverse_closed = false;
//int currentDirection = 0; // 3 = stop, 1 == forward, 2 == reverse 0 == invalid
boolean runningForward = true;
boolean inNeutral = true;

boolean speedControlsChanged = false;
boolean lightingChanged = false;

// when should we turn the headlight off, assuming we want it off?
unsigned long lightOffAt = 0;

// from the command station, at initialization, we get:
unsigned char functionGroupA;
boolean headlightOn = false;

// and, we calculate:
unsigned char encodedSpeedValue = 0;

// power up initalization variables
#define GET_COMMAND_STATION_STATUS_STEP 0
#define GET_COMMAND_STATION_SW_VERSION_STEP 1
#define LOCOMOTIVE_NOT_FREE 2
#define WAIT_FOR_COMMAND_STATION_POWERUP 3
#define GET_LOCOMOTIVE_INFO_STEP 4
#define PAST_INITIALIZATION 200
int initializationStep = GET_COMMAND_STATION_STATUS_STEP;
int initializeError = 0;  // flashes LED pattern on error

// some pre-computed command sequences. Some of the bytes below
// will be computed and put into these byte strings.
unsigned char commandStatusSequence[] = {
  0x21, 0x24, 0x05};
unsigned char commandSWVersionSequence[] = {
  0x21,0x21,0x00};
unsigned char requestAckAck[] = {
  0x20, 0x20};
unsigned char locoInfoRequest[] = {
  0xE3, 0x00, 0x00, 0x00, 0x00};
unsigned char locoSpeedOps[] = {
  0xE4, 0x10, 0x00, 0x00, 0x00, 0x00};
unsigned char locoFunctionSet[] = {
  0xE4, 0x20, 0x00, 0x00, 0x00, 0x00};


// our locomotive, as computed as 2 bytes
unsigned char locomotiveHighByte = 0;
unsigned char locomotiveLowByte = 0;

// speedsteps for 28 speed steps. Choose the values from this table.
// This is quick, easy, and comes directly from the relevant RFP, so it
// should be correct.
unsigned char twentyEightTable[] = {
  B00000, // 0
  B00010, // 1
  B10010, // 2
  B00011, // 3
  B10011, // 4
  B00100, // 5
  B10100, // 6
  B00101, // 7
  B10101, // 8
  B00110, // 9
  B10110, // 10
  B00111, // 11
  B10111, // 12
  B01000, // 13
  B11000, // 14
  B01001, // 15
  B11001, // 16
  B01010, // 17
  B11010, // 18
  B01011, // 19
  B11011, // 20
  B01100, // 21
  B11100, // 22
  B01101, // 23
  B11101, // 24
  B01110, // 25
  B11110, // 26
  B01111, // 27
  B11111 // 28
};

// set up the XpressNet interface on Serial 0 or 1 if we have a machine with more
// than 1 serial port; set it for the only serial port if we only have one.
// Ensure that there are no Serial commands "to the PC" in this case.

void xpressnetSpeed () {    //Set up on 62500 Baud
#ifdef __AVR_ATmega8__
  UBRRH = 0;
  UBRRL = 0x0F;
  UCSRA = 0;
  UCSRB = (1<<RXEN) | (1<<TXEN) | (0<<RXCIE) | (1<<UCSZ2);
  UCSRC = (1<<UCSZ1) | (1<<UCSZ0);
#else
#ifdef SERIAL_PORT_0 // This is the normal port for most Arduinos
  UBRR0H = 0;
  UBRR0L = 0x0F;
  UCSR0A = 0;
  UCSR0B = (1<<RXEN0) | (1<<TXEN0) | (0<<RXCIE0) | (1<<UCSZ02);
  UCSR0C = (1<<UCSZ01) | (1<<UCSZ00);
#else
  UBRR1H = 0;
  UBRR1L = 0x0F;
  UCSR1A = 0;
  UCSR1B = (1<<RXEN1) | (1<<TXEN1) | (0<<RXCIE1) | (1<<UCSZ12);
  UCSR1C = (1<<UCSZ11) | (1<<UCSZ10);
#endif
#endif
}

// calculate the parity bit in the call byte for this guy
unsigned int callByteParity (unsigned int me) {
  int parity = (1==0);
  unsigned int vv;
  me = me & 0x7f;
  vv = me;
  while (vv) {
    parity = !parity;
    vv = vv & (vv-1);
  }
  if (parity) me = me | 0x80;
  return me;
}


void setup() {

  pinMode(8, OUTPUT);        // PARA MI DEBUB A CIEGAS SIN PUERTO SERIE A MAC :-(
#define LED8_ON digitalWrite(8,HIGH);
#define LED8_OFF digitalWrite(8,LOW);

  // the MAX485 chip for Xpressnet - set this up.
  pinMode(MAX485_CONTROL, OUTPUT);
  LISTEN_MODE;

  // turn the LED off until we are running
  pinMode(LED_PIN, OUTPUT);
  LED_OFF;

  // set up the SPDT direction switch. Turn on pullup resistors.
  pinMode(FORWARD_SWITCH, INPUT_PULLUP);
  pinMode(REVERSE_SWITCH, INPUT_PULLUP);
  digitalWrite(FORWARD_SWITCH, HIGH); // pullup resistors ON
  digitalWrite(REVERSE_SWITCH, HIGH); // pullup resistors ON

#ifdef VERBOSE
  delay (2000); // so we can get the monitor open
  Debug.begin (115200);
  Debug.println(" ");
  Debug.println ("Starting setup");
#endif

  dataSaved = 0;

  // calculate the locomotive address
  if (LOCO_ID <= 99) {
    locomotiveHighByte = 0;
    locomotiveLowByte = LOCO_ID;
  } 
  else {
    locomotiveHighByte = (LOCO_ID>>8) + 0xC0;
    locomotiveLowByte = (LOCO_ID&0xFF);
  }

  //  locoInfoRequest string
  locoInfoRequest[2] = locomotiveHighByte;
  locoInfoRequest[3] = locomotiveLowByte;
  calculateXOR ( locoInfoRequest, 5);

  // locoSpeedOps string - put the loco info here
  locoSpeedOps[2] = locomotiveHighByte;
  locoSpeedOps[3] = locomotiveLowByte;

  // locoFunctionSet string - put the loco info here
  locoFunctionSet[2] = locomotiveHighByte;
  locoFunctionSet[3] = locomotiveLowByte;

  // set these up to match call bytes for this device
  myRequestAck = callByteParity (MY_ADDRESS | 0x00) | 0x100;
  myCallByteInquiry = callByteParity (MY_ADDRESS | 0x40) | 0x100;
  myDirectedOps = callByteParity (MY_ADDRESS | 0x60) | 0x100;

  Debug.print ("myAddress is  "); 
  Debug.print (MY_ADDRESS); 
  Debug.print ("  0x"); 
  Debug.println (MY_ADDRESS,HEX);
  Debug.print ("myDirectedOps "); 
  Debug.print ("  0x"); 
  Debug.println (myDirectedOps,HEX);
  Debug.print ("myDirectedWin "); 
  Debug.print ("  0x");  
  Debug.println (myCallByteInquiry,HEX);
  Debug.print ("myRequestAck "); 
  Debug.print ("  0x"); 
  Debug.println (myRequestAck,HEX);

#ifdef VERBOSE
  Debug.println ("finished setup");
#endif

  xpressnetSpeed();   // A PARTIR DE AQUÍ ME QUEDO SIN VER EN EL MAC
}


/* From the ATMega datasheet: */

int USART_Receive( void ) {
  unsigned char status, resh, resl;
  /* Wait for data to be received */
#ifdef __AVR_ATmega8__
  status = UCSRA;
  while ( !(status & (1<<RXC)) ) {
    status = UCSRA;
  }
  /* Get status and 9th bit, then data */
  resh = UCSRB;
  resl = UDR;
  /* If error, return -1 */
  if ( status & ((1<<FE)|(1<<DOR)|(1<<PE))) {
    return -1;
  }
#else
#ifdef SERIAL_PORT_0
  status = UCSR0A;
  //parpadea(5);   hasta aqui bien
  while ( !(status & (1<<RXC0)) ) {
    status = UCSR0A;
  }
  //parpadea(5);  ya no,,,,
  /* Get status and 9th bit, then data */
  resh = UCSR0B;
  resl = UDR0;
  /* If error, return -1 */
  if ( status & ((1<<FE0)|(1<<DOR0)|(1<<UPE0))) { 
    return -1;
  }
#else
  status = UCSR1A;
  while ( !(status & (1<<RXC1)) ) {
    status = UCSR1A;
  }
  /* Get status and 9th bit, then data */
  resh = UCSR1B;
  resl = UDR1;
  /* If error, return -1 */
  if ( status & ((1<<FE1)|(1<<DOR1)|(1<<UPE1))) {
    return -1;
  }
#endif 
#endif

  /* Filter the 9th bit, then return */
  resh = (resh >> 1) & 0x01;
  return ((resh << 8) | resl);
}

void USART_Transmit (unsigned char data8) {
  /* wait for empty transmit buffer */
#ifdef __AVR_ATmega8__
  while (!(UCSRA & (1<<UDRE))) {
  }
  /* put the data into buffer, and send */
  UDR = data8;
#else
#ifdef SERIAL_PORT_0
  while (!(UCSR0A & (1<<UDRE0))) {
  }
  /* put the data into buffer, and send */
  UDR0 = data8;
#else
  while (!(UCSR1A & (1<<UDRE1))) {
  }
  /* put the data into buffer, and send */
  UDR1 = data8;
#endif
#endif
}

// send along a bunch of bytes to the Command Station
void sendCommand(unsigned char *dataString, int byteCount) {
  LED8_ON;
  unsigned int i;
  SEND_MODE;
  delayMicroseconds(20);
  for (i=0; i< byteCount; i++) {
    USART_Transmit (*dataString);
    dataString ++;
  }
  WAIT_FOR_XMIT_COMPLETE;
  LISTEN_MODE;
  LED8_OFF;
}  

// make up the XOR byte, and put it in the LAST position.  
void calculateXOR (unsigned char *dataString, int byteCount) {
  unsigned int i;
  unsigned char xo;
  xo = 0;
  for (i=0; i<(byteCount-1); i++) {
    xo = xo ^ (*dataString);
    dataString ++;
  }
  *dataString = xo;
}

// on a catastrophic error, loop until reset, flashing...
void loopErrorToLED(int pattern) {
  int i;
#ifdef VERBOSE
  Debug.print ("loopErrorToLED, pattern "); 
  Debug.println (pattern);
#endif
  for (;;) {
    for (i=0; i<pattern; i++) {
      LED_ON;
      delay (100);
      LED_OFF;
      delay (100);
    }
    delay (1000);
  }
}


// Our XpressNet Window is open. Do we have anything to send?
void handleCallByteInquiry () { 
  if (initializationStep != PAST_INITIALIZATION) {

    //Debug.print ("handleCallByte, in initialization, step "); Debug.println (initializationStep);

    if ((initializationStep == GET_COMMAND_STATION_STATUS_STEP) ||
      (initializationStep == WAIT_FOR_COMMAND_STATION_POWERUP)) {
      // get the Command Station status 
      sendCommand (commandStatusSequence, 3);    

      // if we are in the waiting for power up, lets assume we are powered up
      // unless told so AGAIN.
      if (initializationStep == WAIT_FOR_COMMAND_STATION_POWERUP) {
        initializationStep = GET_COMMAND_STATION_STATUS_STEP;
      }

    } 
    else if (initializationStep == GET_COMMAND_STATION_SW_VERSION_STEP) {
      // get the Command Station sw version 
      sendCommand (commandSWVersionSequence,3);

    } 
    else if (initializationStep == GET_LOCOMOTIVE_INFO_STEP) {
      // get the info for locomotive in LOCO_ID
      sendCommand (locoInfoRequest,5);

    } 
    else if (initializationStep == LOCOMOTIVE_NOT_FREE) {
      // we have an error - for instance, the locomotive is in an MU, and
      // we are not going to break this MU here.

      loopErrorToLED(initializeError);
    }   
  } 
  else {
    // WE ARE RUNNING, AND OUR WINDOW IS OPEN! YEEEEHAH!
    // anything changed?
    if (speedControlsChanged) {
      sendCommand (locoSpeedOps, 6);
      speedControlsChanged = false;
    } 
    else if (lightingChanged) {
      sendCommand (locoFunctionSet, 6);
      lightingChanged = false;
    } // can put in more else if bits here if there are other things like sound
    // commands to send along.
  }
}

// We have data destined to ALL XpressNet devices; track power off, etc...
void handleGeneralBroadcast () {
  int i;
  if (dataSaved == 4) {
    if (savedData[1] == 0x61) {
      if ((savedData[2] == 0x01) && (savedData[3] == 0x60)) {
        // Normal Operation Resumed
        //Debug.println("Normal Operation Resumed");
        emergencyStop = false;
        trackPowerOff = false;
        inServiceMode = false;
        return;
      } 
      else if ((savedData[2] == 0x00) && (savedData[3] == 0x61)) {
        // Track power off
        // Debug.println("Track power off");
        trackPowerOff = true;
        return;
      } 
      else if ((savedData[2] == 0x02) && (savedData[3] == 0x63)) {
        // Service Mode Entry
        // Debug.println("Service Mode Entry");
        inServiceMode = true;
        return;
      }
    } 
    else if (savedData[1] == 0x81) {
      if ((savedData[2] == 0x00) && (savedData[3] == 0x81)) {
        // Debug.println("Emergency Stop");
        emergencyStop = true;
        return;
      }
    }
  } 
  else if (dataSaved == 8) {
    if (savedData[1] == 0x05) { //DCC FAST CLOCK set request
      //0x05 0xF1 TCODE1 TCODE2 TCODE3 TCODE4 [XOR]
      // Debug.println("DCC FAST CLOCK set request");
      /*
        00mmmmmm	TCODE1, mmmmmm = denotes minutes, range 0...59.
       100HHHHH	TCODE2, HHHHH = denotes hours, range 0...23.
       01000www	TCODE3, www = denotes day of week, 0=monday, 1=tuesday, a.s.o.
       110fffff	TCODE4, fffff = denotes speed ratio, range 0..31. (0=stopped)
       */
      return;
    }
  }

  // There was an error here...
#ifdef VERBOSE
  Debug.print("Broadcast error: ");
  for (i=0; i<dataSaved; i++) {
    Debug.print(savedData[i],HEX); 
    Debug.write(' ');
  }
  Debug.println("");
  dataSaved = 0;
#endif
}

void handleDirectedOps () {  // AQUI PROCESA LOS BYTES DEL BUS, GUARDADOS EN LA MATRIZ savedData[] 
  int i;
  //unsigned char headerNibbleHigh;
  //unsigned char headerNibbleLow;
  unsigned char dataByte1;
  unsigned char dataByte2;
#define DATA_INDICATOR savedData[1]

  if (dataSaved >= 2) {
    //headerNibbleLow = savedData[1] & 0x0f;
    //headerNibbleHigh = savedData[1] & 0xf0;
    dataByte1 = savedData[2];
    dataByte2 = savedData[3];

    if (DATA_INDICATOR == 0x62) {
      // Debug.println("Status");
      if (dataByte1 == 0x21) {    //Version 2.3
        // old version - version 1 and version 2 response.
        loopErrorToLED(5);

      } 
      else if (dataByte1 == 0x22) {
        // Debug.print ("62 com stat indication response data: "); Debug.println (dataByte2, HEX);
        // look at the spec - section 2.1.7 for bits;
        // 0 = 1 - emergency off;
        // 1 = 1 - emergency stop;
        // 2 = 1 - automatic mode, =1, manual mode;
        // 3 = 1 - in service mode;
        // 6 = 1 - in power up;
        // 7 = 1 - RAM check error

        emergencyStop = false;
        trackPowerOff = false;
        inServiceMode = false;

        if (dataByte2 != 0) {
          // is track power turned off?
          if ((dataByte2 & 0x01) == 0x01) {
            trackPowerOff = true;
          }

          // is it in emergency stop?
          if ((dataByte2 & 0x02) == 0x02) {
            emergencyStop = true;
          }

          // in service mode?
          if ((dataByte2 & 0x08) == 0x08) {
            inServiceMode = true;
          }             

          // in powerup mode - wait until complete
          if ((dataByte2 & 0x40) == 0x40) {
            // put us in a state where we do the status request again...
            initializationStep = WAIT_FOR_COMMAND_STATION_POWERUP;
          }
        }
        // if we are powered up, and 
        // We do this as part of the initialization; go to next step
        if (initializationStep == GET_COMMAND_STATION_STATUS_STEP) {
          //Debug.println ("got command station status step, going to sw version");
          initializationStep = GET_COMMAND_STATION_SW_VERSION_STEP;
        }
      } // end of dataByte test

    } 
    else if (DATA_INDICATOR == 0xE2) {
      // Xpressnet 2.1.14.3 - loco info for the MU address...

      // ok - we have info for this guy in initialization.
      if (initializationStep == GET_LOCOMOTIVE_INFO_STEP) {
        initializationStep = LOCOMOTIVE_NOT_FREE;
        initializeError = 2;
      }

    } 
    else if (DATA_INDICATOR  == 0x63) {
      if (dataByte1 == 0x21) {
        // Debug.print ("sw ver, xpressnet "); Debug.print (savedData[2],HEX); Debug.print ("  "); 
        // Debug.print (savedData[3],HEX); Debug.print ("  "); 
        // Debug.println (savedData[4],HEX);


        // we need to be version 3.0 and above; 3.0 is 0x30 in hex.
        if (savedData[3] < 0x30) loopErrorToLED(5);

        // We do this as part of the initialization; go to next step
        if (initializationStep == GET_COMMAND_STATION_SW_VERSION_STEP) {
          //Debug.println ("got sw version, go for locomotive info");
          initializationStep = GET_LOCOMOTIVE_INFO_STEP;
        }
      }

    } 
    else if (DATA_INDICATOR == 0xE3) {
      if (dataByte1 == 0x40) {
        /* Xpressnet 3.6; section 2.1.13.2; "This response is also sent unrequested,
         to the X-Bus device that is currently controlling the locomotive"
         	      Thus, getting this message is ok. */
        // Debug.print ("E3 loco operated by another device ");
        // Debug.print (dataByte2,HEX); Debug.print (" ");
        // Debug.println (savedData[4],HEX);
      }

    }
    else if (DATA_INDICATOR == 0xE4) {   
      // Xpressnet 2.1.14.1 - locomotive info normal locomotive
      // lets look at the loco ID byte:
      if ((dataByte1 & 0x08) == 0x08) {
        // Debug.println ("!! locomotive is being controlled by someone else");
      }

      // Debug.print ("0xE4 -"); Debug.println (dataByte1);
      if ((dataByte1 & 0x07) == 0x00) {
        decoderSpeedSteps = 14; 
        locoSpeedOps[1] = 0x10;
      } 
      else if ((dataByte1 & 0x07) == 0x01) {
        decoderSpeedSteps = 27;
        locoSpeedOps[1] = 0x11;
      } 
      else if ((dataByte1 & 0x07) == 0x02) {
        decoderSpeedSteps = 28;
        locoSpeedOps[1] = 0x12;
      } 
      else if ((dataByte1 & 0x07) == 0x04) {
        decoderSpeedSteps = 128;
        locoSpeedOps[1] = 0x13;
      } 
      else {
        // we don't do this speed steps - something wrong here
        loopErrorToLED(4);           
      }

      // Debug.print ("st: "); Debug.println (decoderSpeedSteps);
      if ((savedData[3] & 0xff) == 0) inNeutral = true;
      if ((savedData[3] & 0x80) == 0x80) runningForward = true; 
      else runningForward = false;

      /* functionGroupB = savedData[5]; */
      /* currentLocoSpeed = savedData[3]; */

      // Save the functions 0-4, so we can adjust them as we see fit; they
      // are in the order 0 0 0 F0  F4 F3 F2 F1
      functionGroupA = savedData[4];
      headlightOn =  ((functionGroupA & 0xF0) == 0x10);      

      // ok - we have info for this guy in initialization.
      if (initializationStep == GET_LOCOMOTIVE_INFO_STEP) {
        initializationStep = PAST_INITIALIZATION; 
      }

    }
    else if (DATA_INDICATOR == 0xE5) {
      // Xpressnet 2.1.14.2 LOCO is part of an MU!

      // ok - we have info for this guy in initialization.
      if (initializationStep == GET_LOCOMOTIVE_INFO_STEP) {
        initializationStep = LOCOMOTIVE_NOT_FREE;
        initializeError = 3;
      }


    }
    else if (DATA_INDICATOR == 0xE6) {

      // Xpressnet 2.1.14.4 - loco info for double header
      // ok - we have info for this guy in initialization.
      if (initializationStep == GET_LOCOMOTIVE_INFO_STEP) {
        initializeError = 1;
        initializationStep = LOCOMOTIVE_NOT_FREE; 
      }
    }
  } // dataSaved >= 2
  // LED8_OFF;
}


void handleSavedData() {    // AQUI DISCRIMINA SI SE HA RECIBIDO MENSAJE GENERAL O PARA NOSOTROS
  // Is this a general broadcast to everyone?
  if (*savedData == GENERAL_BROADCAST) {
    handleGeneralBroadcast();
  } 
  else if (*savedData == myDirectedOps) { // ES PARA NOSOTROS SOLOS?
    handleDirectedOps();
#ifdef VERBOSE  
  } 
  else {
    // else just ignore it for now.
    Debug.write ('?');
#endif
  }
}

// flash the LED if we are in power-off mode      
void handle_possible_track_power_indications() {    
  // what should we do with the LED pin?
  if (trackPowerOff) {
    // let the command station give us our clock; we count
    // "ticks" here.
    currentFlashCounter++;
    if (currentFlashCounter >= flashInterval) {
      currentFlashCounter = 0;

      // flip the state.
      displayLEDState = !displayLEDState;
      if (displayLEDState) {
        LED_ON;
      } 
      else {
        LED_OFF;
      }
    }
  } 
  else {
    // we are in running mode...
    LED_ON;
  }     
}

// Read the potentiometer and switches once every so many xpressnet "ticks"
void read_pots_and_switches() {
  analogueReadCounter ++;
  if (analogueReadCounter == 20) {
    // read the switches for direction change
    int tmp;
    tmp = 0;
    if (digitalRead(FORWARD_SWITCH)) {
      tmp += 1;
    }
    if (digitalRead(REVERSE_SWITCH)) {
      tmp += 2;
    }
    // 3 = neutral, 1 == forward, 2 == reverse 0 == invalid
    switch (tmp) {
      // neutral
    case 0:
    case 3:
      if (!inNeutral) speedControlsChanged = true;
      inNeutral = true;
      break;
      // forward
    case 1:
      inNeutral = false;
      // was last movement reverse?
      if (!runningForward) speedControlsChanged = true;
      runningForward = true;
      break;
    case 2:
      // reverse
      inNeutral = false;
      // was last movement forward?
      if (runningForward) speedControlsChanged = true;
      runningForward = false;
      break;
    }
  } 
  else if (analogueReadCounter == 30) {
    // read the speed for speed changes
    analogueReadCounter = 0;
    potValue = analogRead(POT_PIN);
    potValue = map(potValue,0,1023,0,128);
    // Debug.print ("pv "); Debug.println (potValue);   
    // set to zero speed when direction switch set to neutral
    if (inNeutral) {
      potValue = 0;
    }        
    if (potValue != oldPotValue) {
      // Debug.print ("pot: "); Debug.println (potValue);
      oldPotValue = potValue;
      speedControlsChanged = true;
      lightOffAt = 0;
    }
  }  
  // did the direction switch or pot change?
  if (speedControlsChanged) {
    calculateSpeedDirection();
  }
}

void calculateSpeedDirection() {
  // Debug.print ("direction: "); //Debug.println (currentDirection);
  // Debug.print ("pot: "); Debug.println (potValue);

  // note that the header byte, identification, AH, AL are set previously
  // See Xpressnet 2.2.20.3

  encodedSpeedValue = 0;
  if (decoderSpeedSteps == 28) {
    // from xpressnet 2.1.13.1    
    if (emergencyStop) {
      encodedSpeedValue = 0x01; 
    } 
    else {
      // potValue has been mapped to the range [0-28] via the map function.
      // thus, bounds checking for the following lookup not required.        
      encodedSpeedValue = twentyEightTable[potValue];
    }         
  } 
  else if (decoderSpeedSteps == 128) {
    // from xpressnet 2.1.14.1
    // note that this is from 0 to 127; 127 and 128 we count the same.
    // note that 0 = stop, 1 = emergency stop, that we do not handle yet.
    if (potValue >= 128) {
      encodedSpeedValue = 0x7f; 
    } 
    else encodedSpeedValue = potValue;

    // for now, we do not have an emergency stop button, so we set the
    // speed to zero, and let the locomotive slow down via momentum settings.
    if (encodedSpeedValue == 1) { 
      encodedSpeedValue = 0; 
    }       
  } 
  else if (decoderSpeedSteps == 14) {
    // from xpresset 2.1.12.1
    // 0 = stop
    // 1 = emergency stop
    // 2=15 running values
    if (potValue == 0) { 
      encodedSpeedValue = 0;
    }
    else if (potValue==1) {
      encodedSpeedValue = 0;
    }
    else {
      encodedSpeedValue = potValue + 1;
    }      
  } 
  else if (decoderSpeedSteps == 27) {
    // from xpresset 2.1.12.1
    // 0 = stop
    // 1 = emergency stop
    // 2=15 running values
    if (potValue == 0) { 
      encodedSpeedValue = 0;
    }
    else if (potValue==1) {
      encodedSpeedValue = 0;
    }
    else {
      encodedSpeedValue = potValue + 1;
    }
  } 
  else {
#ifdef VERBOSE
    Debug.print("huh - speed steps"); 
    Debug.println (decoderSpeedSteps);
#endif
    // we do not support this speed step; we should have caught this here
    loopErrorToLED(6);  
  }   
  // do the direction
  if (runningForward) {
    encodedSpeedValue |= 0x80; // set the direction bit
  }                
  // fill in the actual value, and send it along
  locoSpeedOps[4] = encodedSpeedValue;

  // send the speed to the command station
  calculateXOR (locoSpeedOps, 6);
}

// We turn the headlight on when we are moving, off when we are not
void verifyHeadlightIsOK () {
  if (analogueReadCounter == 20) {    
    /* should we ensure that the headlight is on? */
    if (((encodedSpeedValue & 0x7f) != 0) && (!headlightOn)) {
      functionGroupA |= 0x10; //high nibble, bit zero is F0 
      locoFunctionSet[4] = functionGroupA;
      headlightOn = true;
      lightOffAt = 0;
      calculateXOR(locoFunctionSet, 6);
      lightingChanged = true;
      //sendCommand (locoFunctionSet, 6);
    }              
    /* should we ensure that the headlight is off? */
    if (((encodedSpeedValue & 0x7f) == 0) && (headlightOn)) {
      if (lightOffAt == 0) {
        lightOffAt = millis() + LIGHT_OFF_DELAY;
      }
    }  

    if (lightOffAt != 0) {
      if (millis()>lightOffAt) {
        functionGroupA &= 0xEF; //high nibble, bit zero is F0 
        locoFunctionSet[4] = functionGroupA;
        headlightOn = false;
        calculateXOR(locoFunctionSet, 6);
        lightingChanged = true;
        //sendCommand (locoFunctionSet, 6);     
        lightOffAt = 0; 
      }
    }
  }
}

void loop() {
  // get the byte
  int rxdata = USART_Receive();
  // is the data ok?
  if (rxdata != -1) {
    // This IS a Call Byte
    if (rxdata >= 0x100) {
      // is this OUR XpressNet window opening? If so, do we have
      // anything to send?      
      if (rxdata == myCallByteInquiry) {
        handleCallByteInquiry();        
      } 
      else if (rxdata == myRequestAck) {
        sendCommand (requestAckAck,2);
      }       
      if (rxdata == myDirectedOps) {
        // Debug.print(rxdata, HEX);
        // Debug.println("DirOps");
      }      
      // do we have saved data from an old message to me?
      if (dataSaved > 1) {
        handleSavedData();
        dataSaved = 0;
      }
      // is this the start of data coming in for me?  
      if ((rxdata == GENERAL_BROADCAST) ||
        (rxdata == myDirectedOps)) {           
        savedData[0] = rxdata;
        dataSaved=1;       
      } 
      // if not - maybe the call to another window? we just ignore it
    } 
    else {
      // if savedData[0] is for us, we record rest of packet.
      if (dataSaved !=0) {
        // record this data for later processing
        if (dataSaved < MAXSAVED) {
          savedData[dataSaved] = rxdata;
          dataSaved++;
        }
      }
    }    
#ifdef VERBOSE    
  } 
  else {
    // we have a data error...
    //Debug.write("X");
#endif
  }
  // ok - now that we have done this, lets make sure we 
  if (initializationStep == PAST_INITIALIZATION) {   
    handle_possible_track_power_indications();
    read_pots_and_switches(); 
    verifyHeadlightIsOK();
  } 
}


